"""Module atom of viperleed.calc.classes.

Class storing position and other properties of individual atoms (to be
used with Slab, Layer, etc).
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-13'
__license__ = 'GPLv3+'

import copy
import logging

import numpy as np

from viperleed.calc.lib.base import add_edges_and_corners

_LOGGER = logging.getLogger(__name__)


class AtomError(Exception):
    """Base exception for Atom objects."""


class Atom:                                                                     # TODO: Issue #174 -- change description of cartpos when flipping .cartpos[2]
    """Class storing information about a single atom.

    Attributes
    ----------
    el : str
        Element type
    pos : numpy.ndarray
        Atom position as fractional coordinate
    num : int
        A progressive number identifying this atom. Normally, it would
        be the atom number in the POSCAR (i.e., the VESTA progressive,
        non-element-specific number).
    slab : Slab
        The Slab that the atom belongs to.
    layer : Layer or None
        Layer object that the atom belongs to (if any).
    site : Sitetype or None
        Site type of the atom, containing vibrational amplitude and
        occupation. Supplied by the SITE_DEF parameter, assigned by
        the initSites method of Slab.
    cartpos : numpy.ndarray
        Position in Cartesian coordinates, with the highest atom
        at z = 0, positive z pointing into the surface
    linklist : list of Atom
        Defines to which other atoms this Atom is linked. Notice that
        if this Atom is linked to another one, the linklist of both
        are the same object, i.e., there is only one linklist shared
        by all atoms in the same group. This is used together with
        symrefm to determine symmetry-conserving displacements.
    displist : list of Atom
        Like linklist, but keeps track of which symmetry was active
        when displacement was defined.
    freedir : int or numpy.ndarray
        Defines whether the atom can be moved or is locked by symmetry.
        0: no movements, 1: completely free,
        np.array([0|1, 0|1|-1]): parallel movement to a, b, or diagonal
    symrefm : numpy.ndarray
        Defines how a translation of the atom at self.linklist[0]
        should be transformed to affect this atom.
    disp_vib, disp_geo, disp_occ : dict
        Keys are elements, values are lists of vibration/geometry/
        occupation offsets
    disp_geo_offset : dict
        Keys are elements; values are also formatted as lists
        for convenience, but should be one element long.
    disp_center_index : dict of dict
        Which index in the displacement range corresponds to
        'no change'
    dispInitialized : bool
        disp_* variables get initialized after readVIBROCC by Atom.initDisp
    known_deltas : list of str
        Filenames of delta files generated or found for this atom
    offset_geo, offset_vib, offset_occ : dict
        Offsets from self.cartpos, self.site.vib, self.site.occ
        per element
    constraints : dict
        Parameter constraints for restrict.f per element, where
        1, 2, 3 = geo, vib, occ. Can be integer-valued index in
        disp_* range or a tuple (atom, element)
    oriState : Atom
        Deep copy of self before a search is applied
    duplicate_of : Atom
        If this atom is identical to another one by translational
        symmetry (through SYMMETRY_CELL_TRANSFORM or domain supercell
        creation), this points to the atom in the base cell.
    """

    def __init__(self, el, pos, num, slab):
        """Initialize instance."""
        self.el = el
        self.pos = np.asarray(pos, dtype=float)
        self.cartpos = None                                                     # TODO: Issue #174 -- calculate right away
        self.num = num
        self.slab = slab
        self.layer = None
        self.site = None

        self.duplicate_of = None
        self.known_deltas = []
        self.oriState = None

        self.linklist = []
        self.freedir = 1
        self.symrefm = np.identity(2)

        self.displist = []
        self.disp_vib = {'all': [0.]}
        self.disp_geo = {'all': [np.zeros(3)]}
        self.disp_occ = {el: [1.]}
        self.disp_labels = {'geo': '',
                            'vib': '',
                            'occ': ''}
        self.disp_lin_steps = {'geo': [],
                               'vib': [],
                               'occ': []}
        self.disp_geo_offset = {'all': [np.zeros(3)]}
        self.disp_center_index = {'vib': {'all': 0},
                                  'geo': {'all': 0},
                                  'occ': {el: 0}}
        self.dispInitialized = False
        self.offset_geo = {}
        self.offset_vib = {}
        self.offset_occ = {}
        self.constraints = {1: {}, 2: {}, 3: {}}

    def __repr__(self):
        """Return a representation string of this Atom."""
        return (f'{type(self).__name__}(el={self.el}, pos={self.pos}, '
                f'num={self.num}, slab={self.slab})')

    def __str__(self):
        """Return a string version of this Atom."""
        return f'{type(self).__name__}({self.num} {self.el})'

    def __format__(self, fmt_spec):
        """Return a formatted version of self."""
        return format(str(self), fmt_spec)

    @property
    def is_bulk(self):
        """Return whether this Atom is in a bulk layer."""
        try:
            return self.layer.is_bulk
        except AttributeError:
            return False

    def assignConstraint(self, mode, targetel='', value=None, linkAtEl=None,
                         index=None):
        """
        Assign a displacement constraint to this atom. Can be assigned for all
        elements or only one. Constraint is either a fixed value, or another
        (Atom, element) pair to link to.

        Parameters
        ----------
        mode : integer
            Which parameter to constrain. 1: geo, 2: vib, 3: occ
        targetel : string, optional
            If undefined, constrain for all elements. Otherwise,
            only constrain for the given element.
        value : float, optional
            The value to which the given parameter should be
            constrained. If the value is not in the disprange,
            assignment will be skipped. Leave at default (None)
            if 'linkAtEl' or 'index' argument is passed instead.
        linkAtEl : tuple, optional
            Format (Atom, str). The atom and element to which the
            parameter should be linked. If that atom's displist has
            a different length, assignment will be skipped. The
            element may be an empty string. In that case linking
            is considered to apply to all elements. Leave at default
            if 'value' or 'index' argument is passed instead. Default
            is None.
        index : int, optional
            The index to which the given parameter should be
            constrained. Leave at default if 'linkAtEl' or
            'value' argument is passed instead. Indices are
            1-based.

        Raises
        ------
        TypeError
            If no constrain type is given among
            `value`, `linkAtEl`, and `index`
        ValueError
            If more than one constraint type is given among
            `value`, `linkAtEl`, and `index`
        ValueError
            If `mode` is not one of the acceptable displacement modes
        """
        eps = 1e-5
        pars = len([p for p in [value, linkAtEl, index] if p is not None])
        if pars > 1:
            raise ValueError(
                f'{type(self).__name__}.assignConstraint: Can only constrain '
                'to either a fixed value or to another atom, not both'
                )
        if not pars:
            raise TypeError(
                f'{type(self).__name__}.assignConstraint: Exactly one '
                'constraint needed among "index", "value", or "linkAtEl"'
                )

        if mode == 1:
            td = self.disp_geo
        elif mode == 2:
            td = self.disp_vib
        elif mode == 3:
            td = self.disp_occ
        else:  # offset is not allowed here
            raise ValueError(f'{type(self).__name__}.assignConstraint: '
                             f'Unknown key {mode} for mode ({self})')

        if index is not None or value is not None:
            if targetel == '':
                els = list(td.keys())
            else:
                if targetel in td:
                    els = [targetel]
                elif 'all' in td:
                    els = ['all']
                else:
                    _LOGGER.warning(
                        f'Cannot assign constraint for {self}: Element '
                        f'{targetel} does not have displacements assigned.'
                        )
                    return
            for el in els:
                if index:
                    if index > len(td[el]) or index < 1:
                        _LOGGER.warning(
                            f'Cannot assign constraint for {self}, '
                            f'element {el}: index {index} is out of bounds'
                            )
                    else:
                        self.constraints[mode][el] = index - 1
                    continue
                # else value
                if mode == 1:
                    dirvec = td[el][-1] - td[el][0]  # dir of disp range
                    dirvec = dirvec / np.linalg.norm(dirvec)
                    match = [(np.linalg.norm(v - value*dirvec) < eps)
                             for v in td[el]]
                else:
                    match = [(abs(v - value) < eps) for v in td[el]]
                if any(match):
                    self.constraints[mode][el] = match.index(True)
                else:
                    _LOGGER.warning(
                        f'Cannot assign constraint for {self}, element {el}: '
                        f'value {value} is not in displacement list'
                        )
        else:  # linkAtEl
            (at, el2) = linkAtEl
            if mode == 1:
                td2 = at.disp_geo
            elif mode == 2:
                td2 = at.disp_vib
            elif mode == 3:
                td2 = at.disp_occ
            if el2 in td2:
                listlen = len(td2[el2])
            else:
                listlen = len(list(td2.values())[-1])
            if targetel == '':
                els = list(td.keys())
            else:
                els = [targetel]
            for el in els:
                if len(td[el]) != listlen:
                    _LOGGER.warning(f'Cannot constrain {self} to {at}: '
                                    'displacement list lengths differ.')
                    return
            for el in els:
                self.constraints[mode][el] = linkAtEl

    def assignDisp(self, mode, disprange, targetel='', primary=True,
                   displist=[], disp_label='', disp_lin_steps = []):
        """
        Assigns a list of displacements to this atom, for all or a given
        element.

        Parameters
        ----------
        mode : int
            What to displace. 1: geo, 2: vib, 3: occ, 4: geo offset
        disprange :
            The list of displacements. For geometrical offsets, pass
            a list with only one element.
        targetel : str, optional
            If given, assignment is made only for that element,
            otherwise for all. Default is an empty string.
        primary : bool, optional
            Defines whether the assigned displacement should be passed
            along to linked atoms. If True, assignDisp is called for
            these atoms, with primary=False. FOR INTERNAL USE ONLY.
            Default is True.
        displist : list, optional
            Elements are Atom objects. Passed in secondary assignment
            to later link parameters (the 'linklist' defines how the
            'displist' is defined, but can change via the SYM_DELTA
            parameter).

        Raises
        ------
        ValueError
            If `mode` is not an acceptable displacement mode.
        """
        if targetel.lower() == 'vac':
            # don't write stuff for vacancies - they don't get displaced, and
            #   occupation can be determined implicitly
            return
        eps = 1e-5  # tolerance for comparing displacements
        dr = copy.copy(disprange)
        # to make sure multiple atoms do not get the same list object
        if mode == 1:
            td = self.disp_geo
            self.disp_labels['geo'] = disp_label
            self.disp_lin_steps['geo'] = disp_lin_steps
        elif mode == 2:
            td = self.disp_vib
            self.disp_labels['vib'] = 'N/A'  # direction not applicable for vib
            self.disp_lin_steps['vib'] = disp_lin_steps
        elif mode == 3:
            td = self.disp_occ
            self.disp_labels['occ'] = 'N/A'  # direction not applicabel for occ
            self.disp_lin_steps['occ'] = disp_lin_steps
        elif mode == 4:
            td = self.disp_geo_offset
        else:
            raise ValueError(f'{type(self).__name__}.assignDisp: '
                             f'Unknown key {mode} for mode ({self})')
        if targetel == '':
            els = list(td.keys())
        else:
            els = [targetel]
        for el in els:
            if mode == 4:  # special case: offset
                if el not in td:
                    td[el] = dr
                else:
                    match = True
                    if (abs(dr[0][2]) > eps and
                            abs(td[el][0][2] - dr[0][2]) > eps):
                        if abs(td[el][0][2]) < eps:
                            td[el][0][2] = dr[0][2]
                        else:
                            match = False
                    if (np.linalg.norm(dr[0][:2]) > eps and
                            np.linalg.norm(td[el][0][:2] - dr[0][:2]) > eps):
                        if all([(abs(f) < eps) for f in td[el][0][:2]]):
                            td[el][0][:2] = dr[0][:2]
                        else:
                            match = False
                if not match:
                    _LOGGER.warning(
                        'Atom.assignDisp: Trying to assign offset, '
                        f'but {self} already has an offset defined.'
                        ' Skipping second assignment.'
                        )
                continue
            if (el not in td
                    or (len(td[el]) == 1 and np.linalg.norm(td[el]) < 1e-5)
                    or (mode == 3 and len(td[el]) == 1)):
                # did not assign for this element yet -> OK, store
                td[el] = dr
                # also store center
                if mode != 3:
                    n = [np.linalg.norm(v) for v in dr]
                else:
                    n = [abs(v - self.site.occ[el]) for v in dr]
                smode = {1: 'geo', 2: 'vib', 3: 'occ'}
                self.disp_center_index[smode[mode]][el] = n.index(min(n))
                continue
            # is warning required: every value in td[el] also in disprange?
            match = True
            if el in td:
                if len(td[el]) != len(dr):
                    match = False
                else:
                    for v in td[el]:
                        # check whether all values from td[el] are in disprange
                        found = False
                        for v2 in dr:
                            if np.linalg.norm(v-v2) < eps:
                                found = True
                                break
                        if not found:
                            match = False
                            break
                if not match and len(td[el]) > 1:
                    _LOGGER.warning(
                        'Atom.assignDisp: Trying to assign displacement list, '
                        f'but {self} already has displacements assigned. '
                        'Skipping second assignment.'
                        )
                    return    # also don't assign to other atoms
                # in case of linking, base assignments on current dr
                dr = td[el][:]
        # assign to atoms in linklist:
        if primary:
            if not self.displist:
                self.displist = [self]
                self.slab.displists.append(self.displist)
            for at in [at for at in self.linklist if at != self]:
                if mode == 1 or mode == 4:
                    tm = np.identity(3)
                    tm[:2, :2] = np.dot(at.symrefm,
                                        np.linalg.inv(self.symrefm))
                    newdr = [np.dot(tm, v) for v in dr]
                else:
                    newdr = dr[:]
                at.assignDisp(mode, newdr, targetel, primary=False,
                              displist=self.displist,
                              disp_label=disp_label,
                              disp_lin_steps=disp_lin_steps)
            return
        if self.displist and displist != self.displist:
            _LOGGER.warning(
                f'{self} is being linked to different groups in the '
                'DISPLACEMENTS file. This will interfere with correct '
                'parameter linking in the search! Check SYM_DELTA settings.'
                )
        if self not in displist:
            displist.append(self)
        self.displist = displist

    def clearOffset(self, mode, targetel='', primary=True, displist=[]):
        """Revert an offset for self and all its symmetry-equivalent atoms.

        The offset restored is the one saved in .oriState
        (typically from POSCAR or VIBROCC).

        Parameters
        ----------
        mode : int
            Which offset to restore. 1: geo, 2: vib, 3: occ
        targetel : str, optional
            If passed, assignment is made only for that element,
            otherwise for all.
        primary : bool, optional
            Defines whether assignment should be passed along
            to linked atoms. This will call assignDisp for these
            atoms, with primary=False.
        displist : list, optional
            Elements are Atom objects. Passed in secondary assignment
            to later link parameters (the 'linklist' defines how the
            'displist' is defined, but can change via the SYM_DELTA
            parameter).

        Raises
        ------
        ValueError
            If `mode` is not one of the acceptable displacement modes.
        """
        if self.oriState is None or targetel.lower() == 'vac':
            return
        if mode == 1:
            td = self.offset_geo
            od = self.oriState.offset_geo
        elif mode == 2:
            td = self.offset_vib
            od = self.oriState.offset_vib
        elif mode == 3:
            td = self.offset_occ
            od = self.oriState.offset_occ
        else:
            raise ValueError(f'{type(self).__name__}.clearOffset: '
                             f'Unknown key {mode} for mode ({self})')
        if targetel == '':
            els = list(td.keys())
        else:
            els = [targetel]
        for el in els:
            if el not in od:
                del td[el]
            else:
                td[el] = od[el]
        # assign to atoms in linklist:
        if primary:
            if not self.displist:
                self.displist = [self]
                self.slab.displists.append(self.displist)
            for at in [at for at in self.linklist if at != self]:
                at.clearOffset(mode, targetel, primary=False,
                               displist=self.displist)
            return
        if self.displist and displist != self.displist:
            _LOGGER.warning(
                f'{self} is being linked to different groups in the '
                'DISPLACEMENTS file. This will interfere with correct '
                'parameter linking in the search! Check SYM_DELTA settings.'
                )
        if self not in displist:
            displist.append(self)
        self.displist = displist

    def copyOriState(self, other):
        """Deepcopy positions and offsets from another atom into oriState."""
        self.storeOriState()
        self.oriState.pos = other.pos.copy()
        self.oriState.cartpos = other.cartpos.copy()
        self.oriState.offset_geo = copy.deepcopy(other.offset_geo)
        self.oriState.offset_vib = copy.deepcopy(other.offset_vib)
        self.oriState.offset_occ = copy.deepcopy(other.offset_occ)

    def duplicate(self, add_to_atlists=True, num=None):
        """Return a somewhat lightweight copy of this Atom.

        Attributes position and elements are deep-copied, all others
        are instead references to those of this Atom. This includes
        in particular, site, displacements, slab, and layer. The new
        Atom can also be automatically added to the existing slab and
        layer.

        Parameters
        ----------
        add_to_atlists : bool, optional
            Whether the duplicate atom should be added to atom lists
            in the existing Slab and Layer objects. Default is True.
        num : int or None, optional
            Progressive number to give to the new atom returned.
            If not given or None, it is taken from the slab of
            this Atom. Default is None.

        Returns
        -------
        newat : Atom
            The duplicate atom that was created.
        """
        if num is None:
            num = self.slab.n_atoms + 1
        newat = Atom(self.el, self.pos.copy(), num, self.slab)
        newat.cartpos = self.cartpos.copy()
        newat.duplicate_of = self
        newat.site = self.site
        newat.dispInitialized = True
        newat.disp_vib = self.disp_vib
        newat.disp_geo = self.disp_geo
        newat.disp_occ = self.disp_occ

        if add_to_atlists:
            self.slab.atlist.append(newat)                                      # TODO: consider an AtomContainer.add_atom(atom) abstract method!
            self.slab.n_per_elem[self.el] += 1
            if self.layer is not None:
                self.layer.atlist.append(newat)
                newat.layer = self.layer
        return newat

    def initDisp(self, force=False):                                            # TODO: all the DISP stuff should complain if .is_bulk
        """Initialize displacement dictionaries based on self.site.

        This method should not be called before a site has been
        given to this Atom.

        This method must be called before displacements can be
        merged with their offsets, as this method prepares the
        correct entries in disp_occ.

        Parameters
        ----------
        force : bool, optional
            Whether displacements should be cleared also
            if they are already present. Default is False.

        Raises
        ------
        RuntimeError
            If this method is called before a site is available
        """
        if not self.site:
            raise RuntimeError('Cannot initialize displacements '
                               'before a site is defined')
        if self.dispInitialized and not force:
            return

        self.dispInitialized = True
        self.disp_vib = {'all': [0.]}
        self.disp_geo = {'all': [np.zeros(3)]}
        self.disp_occ = {}
        self.disp_center_index = {'vib': {'all': 0},
                                  'geo': {'all': 0},
                                  'occ': {}}
        for k, v in self.site.occ.items():
            if v > 0 or k in self.site.mixedEls:
                self.disp_occ[k] = [v]
                self.disp_center_index['occ'][k] = 0

    def is_same_xy(self, cartpos, eps=1e-3):
        """Return whether this atom is close to a 2D cartpos.

        If the atom is close to an edge or corner its replicas
        are also considered.

        Parameters
        ----------
        cartpos : numpy.ndarray or Atom
            2D Cartesian coordinates to check against the position
            of this atom. If an Atom, its in-plane Cartesian position
            is used.
        eps : float, optional
            The precision to which positions are expected to match.
            The default is 1e-3.

        Returns
        -------
        bool
            True if positions match, else False.
        """
        if isinstance(cartpos, Atom):
            cartpos = cartpos.cartpos[:2]
        abt = self.slab.ab_cell.T
        releps = eps / np.linalg.norm(abt, axis=1)
        complist, _ = add_edges_and_corners([self.cartpos[:2]], (self.pos,),
                                            releps, abt)
        return any(np.linalg.norm(cartpos - complist, axis=1) < eps)

    def mergeDisp(self, el):
        """
        Merges the offsets from VIBROCC and DISPLACEMENTS into the
        displacements lists from DISPLACEMENTS for the given element.

        For vibrational and occupational offsets a consistency check is
        performed. The offset lists will be emptied.

        Raises
        -------
        RuntimeError
            If this method is called before initDisp.
        """
        if not self.dispInitialized:
            raise RuntimeError('Need to .initDisp before displacements '
                               'can be merged with offsets')

        self.storeOriState()
        # geometric offsets from DISPLACEMENTS
        geo_d_offset = self.disp_geo_offset.get(el,
                                          self.disp_geo_offset['all'])[0]
        if el not in self.disp_geo:
            self.disp_geo[el] = copy.copy(list(self.disp_geo['all']))
        self.disp_geo[el] = list(self.disp_geo[el] + geo_d_offset)
        self.disp_geo_offset = {'all': [np.zeros(3)]}

        # geometric offsets from VIBROCC
        if el in self.offset_geo:
            geo_offset = self.offset_geo[el]
            self.disp_geo[el] = [geo_step + geo_offset for geo_step in self.disp_geo[el]]
            del self.offset_geo[el]

        # vibrational offsets from VIBROCC
        if el not in self.disp_vib:
            self.disp_vib[el] = copy.copy(self.disp_vib['all'])
        if el in self.offset_vib:
            vib_offset = self.offset_vib[el]
            final_vib_steps = [vib_step + vib_offset for vib_step in self.disp_vib[el]]
            if any(np.array(final_vib_steps) + self.site.vibamp[el] < 0):
                _LOGGER.error(f'Vibrational offset for {self} defined in '
                              'VIBROCC would result in negative vibrational '
                              'amplitude. Offset will be ignored.')
            else:
                self.disp_vib[el] = final_vib_steps
            del self.offset_vib[el]

        # vibrational offsets from VIBROCC
        if el in self.offset_occ:
            occ_offset = self.offset_occ[el]
            final_occ_steps = [occ_step + occ_offset for occ_step in self.disp_occ[el]]
            if any(np.array(final_occ_steps) < 0) or any(np.array(final_occ_steps) > 1):
                _LOGGER.error(
                    f'Occupational offset for {self} defined in '
                    'VIBROCC would result in unphysical concentration'
                    '(occupation <0 or >1). Offset will be ignored.'
                    )
            else:
                self.disp_occ[el] = final_occ_steps
            del self.offset_occ[el]

    def storeOriState(self):
        """Stores the initial values from the input files for this atom."""
        if self.oriState is None:
            self.oriState = self.duplicate(add_to_atlists=False)                # TODO: potential problem: duplicate only shallow copies some attributes.

    def translate_2d(self, cart_shift, frac_shift):
        """Apply a 2D translation to this Atom."""
        self.cartpos[:2] += cart_shift
        self.pos[:2] += frac_shift



"""Functions for reading and writing the VIBROCC file."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging
import re

import numpy as np

from viperleed.calc.lib.base import readToExc
from viperleed.calc.lib.base import splitSublists


logger = logging.getLogger(__name__)


def readVIBROCC(rp, slab, filename='VIBROCC', silent=False):
    """Reads VIBROCC and adds the information to all sites in the slab.
    If vibrational amplitudes are automatically calculated here, returns True,
    else returns False."""
    if not (rp.T_EXPERIMENT is None or rp.T_DEBYE is None):
        generate = True
    else:
        generate = False
    vibAmpGenerated = False
    # open input file
    try:
        rf = open(filename, 'r')
    except FileNotFoundError:
        if generate:
            logger.info("No VIBROCC file found, generating vibrational "
                        "amplitudes from temperature...")
            for site in slab.sitelist:
                if site.el not in rp.ELEMENT_MIX:
                    if site.getVibAmp(rp, site.el) is not None:
                        logger.error("Failed to generate vibrational "
                                     "amplitude for " + site.el + " in site "
                                     + site.label)
                        raise
                else:
                    for subel in rp.ELEMENT_MIX[site.el]:
                        if site.getVibAmp(rp, subel) is not None:
                            logger.error(
                                "Failed to generate vibrational amplitude for "
                                + site.el + " in site " + site.label)
                            raise
            try:
                checkVIBROCC(rp, slab, generate=True)
            except Exception:
                raise
            return True
        else:
            logger.error("VIBROCC not found.")
            raise
    # read VIBROCC:
    mode = 0  # 0: not reading; 1: vib.amp., 2: occ., 3: search offsets
    regex = False   # read regular expressions as-is or not
    for line in rf:
        line = line.lstrip()
        for c in ["!", "#", "%"]:    # start of comment
            line = line.split(c)[0].rstrip()
        if '=' not in line:
            continue
        else:
            if line[0] == '=':
                llist = line[1:].split()
                if llist[0][0].lower() == 'v':
                    mode = 1
                elif llist[0][0].lower() == 'o':
                    mode = 2
                elif llist[0][0].lower() == 's':
                    mode = 3
                elif llist[0][0].lower() == 'r':
                    regex = True
                    if len(llist) >= 2:
                        if llist[1].lower() == 'off':
                            regex = False
                    continue
                else:
                    logger.warning("VIBROCC: Found line starting with '=', "
                                   "but didn't recognize vibrational "
                                   "amplitude or occupation")
                    rp.setHaltingLevel(1)
                continue
            elif mode == 0:
                continue
            else:
                param = line.split('=')[0]
                if not param:
                    continue
                else:
                    plist = param.split()
                    if plist:
                        param = plist[0]
                    else:
                        continue
        if mode in [1, 2]:
            try:
                llist = line.split('=')[1].split()
            except IndexError:
                logger.warning(f'VIBROCC file: {param} appears to have '
                               'no value')
                rp.setHaltingLevel(1)
                continue
            if not llist:
                logger.warning(f'VIBROCC file: {param} appears to have '
                               'no value')
                rp.setHaltingLevel(1)
                continue
            # first get values on the right
            sublists = splitSublists(readToExc(llist), ',')
            # read parameter on the left
            targetsites = []
            for site in slab.sitelist:
                if regex:        # if regular expressions are enabled, take the
                    prep = param     # parameter input at face value
                else:
                    # double-slash non-literal characters
                    prep = re.escape(param)
                    # if regular expressions are not enabled, we want to still
                    #  interpret * as "any number of any characters":
                    prep = prep.replace('\\*', '.*')
                m = re.match(prep, site.label)
                if m:
                    # if the length of the matched text == the site label,
                    #  it's a real match
                    if m.end(0) == len(site.label):
                        targetsites.append(site)
            if len(targetsites) == 0:
                logger.warning('VIBROCC file: No sites matching '+param
                               + ' found, line will be skipped.')
                rp.setHaltingLevel(1)
                continue
            else:
                for site in targetsites:
                    if mode == 1:   # decide which dictionary to write to
                        td = site.vibamp
                    else:
                        td = site.occ
                    for sl in sublists:
                        if len(sl) == 1:
                            # if there's only one value, it should be a float
                            #  for the main site element
                            try:
                                if site.el in rp.ELEMENT_MIX:
                                    for subel in rp.ELEMENT_MIX[site.el]:
                                        td[subel] = float(sl[0])
                                else:
                                    td[site.el] = float(sl[0])
                            except Exception:
                                logger.error(
                                    'VIBROCC file: Error reading '
                                    'value '+sl[0]+' at parameter '+param)
                                raise
                        else:
                            el = sl[0].capitalize()
                            try:
                                td[el] = float(sl[1])
                            except KeyError:
                                logger.error(
                                    'VIBROCC file: Element ' + el + 'not '
                                    'recognized at parameter ' + param)
                            except Exception:
                                logger.error(
                                    'VIBROCC file: Error reading value '
                                    + sl[1] + ' at parameter ' + param)
                                raise
        # Search Offsets
        if mode == 3:
            try:
                ind = int(plist[1])
            except Exception:
                logger.error('VIBROCC file: Error converting to atom number: '
                             + plist[1])
                continue
            else:
                targetatlist = [at for at in slab if at.num == ind]
                if len(targetatlist) == 1:
                    targetat = targetatlist[0]
                else:
                    logger.error('VIBROCC file: {} atoms found with number {}'
                                 .format(len(targetatlist), ind))
                    continue
            if param.lower() in ["pos", "geo"]:
                om = 1
                targetdict = targetat.offset_geo
            elif param.lower() == "vib":
                om = 2
                targetdict = targetat.offset_vib
            elif param.lower() == "occ":
                om = 3
                targetdict = targetat.offset_occ
            else:
                logger.error('VIBROCC file: Flag not recognized: '+param)
                continue
            s = line.split("=")[1]
            sub_lists = s.split(",")
            for li in sub_lists:
                ll = li.split()
                if (len(ll) != 4 and om == 1) or (len(ll) != 2 and om != 1):
                    logger.error('VIBROCC file: Wrong number of values in '
                                 'sublist '+li)
                    continue
                else:
                    el = ll[0].capitalize()
                    if el not in slab.chemelem:
                        logger.error('VIBROCC file: Element '+el+' not '
                                     'recognized')
                        continue
                    values = []
                    try:
                        for s in ll[1:]:
                            values.append(float(s))
                    except ValueError:
                        logger.error('VIBROCC file: Could not convert values '
                                     'to floats: '+li)
                        continue
            if om == 1:
                value = np.array(values)
                value[2] *= -1  # invert z
            else:
                value = values[0]
            targetdict[el] = value
    rf.close()
    if not silent:
        logger.debug("VIBROCC file was read successfully")
    # now fill up default values & do consistency checks:
    vibAmpGenerated = checkVIBROCC(rp, slab, generate=generate, silent=silent)
    if vibAmpGenerated:
        return True
    else:
        if rp.T_EXPERIMENT is not None:
            logger.warning("Parameter T_EXPERIMENT is defined but unused.")
        if rp.T_DEBYE is not None:
            logger.warning("Parameter T_DEBYE is defined but unused.")
    return False


def checkVIBROCC(rp, slab, generate=False, silent=False):
    """Fills default values and does consistency check of site vibrational
    amplitudes and occupations. If vibrational amplitudes are automatically
    calculated here, returns True, else returns False."""
    vibAmpGenerated = False
    for site in slab.sitelist:
        # check if the vibrational amplitude is defined for the main element(s)
        if site.el not in site.vibamp:
            if site.el not in rp.ELEMENT_MIX:
                if generate:
                    if site.getVibAmp(rp, site.el) == 0:
                        vibAmpGenerated = True
                    else:
                        logger.warning(
                            'VIBROCC file: Failed to get ' + site.el +
                            ' vibrational amplitude for site ' + site.label)
                        rp.setHaltingLevel(2)
                else:
                    logger.warning(
                        'VIBROCC file: No ' + site.el + ' vibrational '
                        'amplitude defined for site '+site.label)
                    rp.setHaltingLevel(2)
            else:
                v = -1
                for subel in rp.ELEMENT_MIX[site.el]:
                    if subel in site.vibamp:
                        v = site.vibamp[subel]
                    elif generate:
                        if site.getVibAmp(rp, subel) == 0:
                            vibAmpGenerated = True
                            v = site.vibamp[subel]
                if v == -1:
                    logger.error(
                        'VIBROCC file: No vibrational amplitude defined for '
                        'any of the main elements in site ' + site.label)
                    rp.setHaltingLevel(3)
                    # !!! raise exception
                else:
                    # vibrational amplitudes were defined for some of the
                    #  main elements but not all. Fill the other values
                    for subel in rp.ELEMENT_MIX[site.el]:
                        if subel not in site.vibamp:
                            logger.warning(
                                'VIBROCC file: No ' + subel + ' vibrational '
                                'amplitude defined for site ' + site.label +
                                '. Using vibrational amplitude from other '
                                'element in site.')
                            rp.setHaltingLevel(1)
                            site.vibamp[subel] = v
        if site.el in rp.ELEMENT_MIX:  # just use first element in ELEMENT_MIX
            mainva = site.vibamp[rp.ELEMENT_MIX[site.el][0]]
        else:
            mainva = site.vibamp[site.el]
        # for other elements, fill vibrational elements with that of the main
        # element (probably not present)
        for el in slab.chemelem:
            if el not in site.vibamp:
                site.vibamp[el] = mainva
        # check and fill occupations:
        o = 0.
        for el in slab.chemelem:
            if el in site.occ:
                o += site.occ[el]
        if 'Vac' in site.occ:
            o += site.occ['Vac']
        if site.el not in site.occ:
            if site.el in rp.ELEMENT_MIX:
                if not any([(el in site.occ)
                            for el in rp.ELEMENT_MIX[site.el]]):
                    for el in rp.ELEMENT_MIX[site.el]:
                        site.occ[el] = (1. - o) / len(rp.ELEMENT_MIX[site.el])
            else:
                site.occ[site.el] = 1. - o
            o = 1.
        for el in slab.chemelem:
            if el not in site.occ:
                site.occ[el] = 0.
        if o < 0.99:
            logger.debug(
                'VIBROCC file: Site '+site.label+' has a total '
                'occupation less than one. Interpreting remaining {:.2f} '
                'as vacancies.'.format(1-o))
        elif o > 1.0:
            if o > 1.01:
                logger.warning(
                    'VIBROCC file: Site '+site.label+' has a total occupation '
                    'greater than one ({:.2f}). Occupations will be re-scaled '
                    'to 1.'.format(o))
                rp.setHaltingLevel(2)
            for el in site.occ:
                site.occ[el] *= 1/o
        # finally, check whether we have any vibrational amplitudes or
        #  occupations assigned to non-existant elements, and if so, drop
        #  them and warn the user about it:
        dl = []
        for el in site.vibamp:
            if el not in slab.chemelem:
                logger.warning(
                    'VIBROCC file: Site ' + site.label + ' has a vibrational '
                    'amplitude defined for an unknown element, which will be '
                    'dropped (' + el + ').')
                rp.setHaltingLevel(1)
                dl.append(el)
        for el in dl:
            site.vibamp.pop(el, None)
        dl = []
        for el in site.occ:
            if el not in slab.chemelem and el != "Vac":
                logger.warning(
                    'VIBROCC file: Site ' + site.label + ' has an occupation '
                    'defined for an unknown element, which will be dropped ('
                    + el + ').')
                rp.setHaltingLevel(1)
                dl.append(el)
        for el in dl:
            site.occ.pop(el, None)
    if not silent:
        logger.debug("VIBROCC value consistency check finished")
    return vibAmpGenerated


def writeVIBROCC(sl, rp, filename="VIBROCC_OUT", silent=False):
    """Writes a new VIBROCC file with the optimized parameters obtained after
    the search. The new file will not follow the ordering of the input VIBROCC
    file."""
    output = "= Vibrational Amplitudes\n"
    for site in sl.sitelist:
        ol = site.label + " = "
        targetels = [el for el in site.vibamp if (site.occ[el] != 0
                                                  or el in site.mixedEls)]
        if len(targetels) == 1:
            ol += "{:.4g}".format(site.vibamp[targetels[0]])
        else:
            for i, el in enumerate(targetels):
                ol += el + " {:.4g}".format(site.vibamp[el])
                if i < len(targetels)-1:
                    ol += ", "
        output += ol + "\n"

    output += "\n= Occupations\n"
    for site in sl.sitelist:
        write = True
        ol = site.label + " = "
        targetels = [el for el in site.occ if site.occ[el] != 0]
        if (len(targetels) == 1 and site.occ[targetels[0]] == 1
                and site.el == targetels[0] and not site.mixedEls):
            write = False
        elif len(targetels) == 1 and not site.mixedEls:
            ol += "{:.4g}".format(site.occ[targetels[0]])
        else:
            for i, el in enumerate(targetels):
                ol += el + " {:.4g}".format(site.occ[el])
                if i < len(targetels)-1:
                    ol += ", "
        if write:
            output += ol + "\n"

    output += "\n= Search offsets\n"
    # figure out which atoms to write, in which order
    offsetList = []    # metric per atom
    for at in [at for at in sl if not at.is_bulk]:
        to = 0
        # !!! TODO: Think about weights for the three
        for el in at.offset_occ:
            to += abs(at.offset_occ[el])
        for el in at.offset_vib:
            to += abs(at.offset_vib[el])
        for el in at.offset_geo:
            to += np.linalg.norm(at.offset_geo[el])
        if to >= 1e-4:  # !!! TODO: enough?
            offsetList.append((to, at))
    offsetList.sort(key=lambda tup: -tup[0])
    for (_, at) in offsetList:
        # POSITION OFFSET
        write = False
        ol = "POS {} = ".format(at.num)
        for el in at.offset_geo:
            if np.linalg.norm(at.offset_geo[el]) >= 1e-4:
                write = True
                ol += el + " {:g} {:g} {:g}, ".format(at.offset_geo[el][0],
                                                      at.offset_geo[el][1],
                                                      -at.offset_geo[el][2])
        if write:
            output += ol[:-2]+"\n"
        # VIBRATION OFFSET
        write = False
        ol = "VIB {} = ".format(at.num)
        for el in at.offset_vib:
            if abs(at.offset_vib[el]) >= 1e-3:
                write = True
                ol += el + " {:g}, ".format(at.offset_vib[el])
        if write:
            output += ol[:-2]+"\n"
        # OCCUPATION OFFSET
        write = False
        ol = "OCC {} = ".format(at.num)
        for el in at.offset_occ:
            if abs(at.offset_occ[el]) >= 1e-3:
                write = True
                ol += el + " {:g}, ".format(at.offset_occ[el])
        if write:
            output += ol[:-2]+"\n"
    try:
        with open(filename, 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write "+filename)
        raise
    if not silent:
        logger.info("Wrote to "+filename+" successfully")
    return



"""Functions for writing the SearchProgress.pdf and SearchReport.pdf files."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging

from matplotlib.markers import MarkerStyle
import numpy as np

try:
    import matplotlib
    matplotlib.rcParams.update({'figure.max_open_warning': 0})
    matplotlib.use('Agg')  # !!! check with Michele if this causes conflicts
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
    plt.style.use('viperleed.calc')
except Exception:
    _CAN_PLOT = False
else:
    _CAN_PLOT = True


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def writeSearchProgressPdf(rp, gens, rfacs, lastconfig,
                           outname="Search-progress.pdf",
                           csvname="Search-progress.csv",
                           markers=None,
                           rfac_predict=None):
    """
    Writes a pdf file with reports on R-factor convergence and current
    parameter scatter. Also writes a csv file containing the most basic
    information.

    Parameters
    ----------
    rp : Rparams
        The run parameters
    gens : list of int
        List of generation numbers for which data is stored.
    rfacs : list of (list of float)
        The R-factors for a given generation (corresponding to gens)
    lastconfig : tuple
        Lists (percent, dc) for each domain, with dc the parameter values of
        that domain
    outname : str, optional
        The file name to write to. The default is "Search-progress.pdf".
    csvname : str, optional
        The file name of the csv file to write to. The default is
        "Search-progress.csv".
    markers : list of tuples (int, str), optional
        Pass x-positions in the generations plots to be marked by vertical
        lines as tuples (gen, label), where gen is the position in the plot.
    rfac_predict : list of float, optional
        List of r-factor values determined from the parabola fit, by
        generation. The default is [].

    Returns
    -------
    None.

    """
    global _CAN_PLOT
    if not _CAN_PLOT:
        return None

    markers = markers or []

    figsPerPage = 5
    parsPerFig = 8

    searchname = rp.disp_blocks[rp.search_index][1]

    rfacsMin = np.array([min(rfa) for rfa in rfacs])
    rfacsMax = np.array([max(rfa) for rfa in rfacs])
    rfacsMean = np.array([np.mean(rfa) for rfa in rfacs])
    rp.searchplots[-1] = (searchname, gens, rfacsMin, rfacsMax, rfacsMean)
    # rfacsStd = np.array([np.std(rfa, ddof = 1) for rfa in rfacs])
    rlastunique = [rfacs[-1][0]]
    lastpops = [1]
    for r in rfacs[-1][1:]:
        if r == rlastunique[-1]:
            lastpops[-1] += 1
        else:
            rlastunique.append(r)
            lastpops.append(1)
    allcolors = []
    if rlastunique[-1] == rlastunique[0]:
        colors = [(0., 0., 0.)] * len(rlastunique)  # black
        allcolors = [(0., 0., 0.)] * len(rfacs[-1])
    else:
        colors = []
        for (i, r) in enumerate(rlastunique):
            w = (r - rlastunique[0]) / (rlastunique[-1] - rlastunique[0])
            w = np.sqrt(w)   # stronger scaling towards red
            colors.append((w, 0., 0.))
            for j in range(0, lastpops[i]):
                allcolors.append((w, 0., 0.))
    if (not rp.rfacscatter_all) or (rp.rfacscatter_all[-1][0] != gens[-1]):
        rp.storeRfacScatter([gens[-1]]*len(rlastunique), rlastunique,
                            lastpops, colors)
    deltagens = {"all": {1: 0}, "best": {1: 0}, "dec": {1: 0}}
    previous = {"all": rfacs[0][:],
                "best": min(rfacs[0]),
                "dec": rfacs[0][:(len(rfacs) // 10 + 1)]}
    for i in range(1, len(gens)):
        if min(rfacs[i]) != previous["best"]:
            deltagens["best"][gens[i]] = (gens[i]
                                          - max(deltagens["best"].keys()))
            previous["best"] = min(rfacs[i])
        if list(rfacs[i]) != list(previous["all"]):
            deltagens["all"][gens[i]] = (gens[i]
                                         - max(deltagens["all"].keys()))
            previous["all"] = rfacs[i][:]
        if list(rfacs[i][:(len(rfacs[i]) // 10 + 1)]) != list(previous["dec"]):
            deltagens["dec"][gens[i]] = (gens[i]
                                         - max(deltagens["dec"].keys()))
            previous["dec"] = rfacs[i][:(len(rfacs[i]) // 10 + 1)]

    figsize = (5.8, 8.3)
    figs = []

    # CSV output
    sep = "; "
    width = 12
    titles = ["Generation", "G_Delta_all", "G_Delta_dec", "G_Delta_best",
              "R_min", "R_max", "R_mean"]
    output = ""
    for t in titles:
        output += t.rjust(width)+sep
    output = output[:-len(sep)] + "\n"
    mc = markers[:]
    for i in range(0, len(gens)):
        if mc:
            if gens[i] > mc[0][0]:
                output += mc[0][1] + "\n"   # comment line for marker
                mc.pop(0)
        output += str(gens[i]).rjust(width) + sep
        for k in ["all", "dec", "best"]:
            if gens[i] in deltagens[k]:
                output += str(deltagens[k][gens[i]]).rjust(width) + sep
            else:
                output += "NaN".rjust(width) + sep
        for ls in [rfacsMin, rfacsMax, rfacsMean]:
            output += "{:.4f}".format(ls[i]).rjust(width)+sep
        output = output[:-len(sep)] + "\n"
    try:
        with open(csvname, "w") as wf:
            wf.write(output)
    except KeyboardInterrupt:
        raise
    except Exception:
        logger.warning("Failed to write "+csvname)

    # R-FACTOR AND GENERATION DELTA
    # create figure
    fig, (rfp, dgp) = plt.subplots(2, 1, sharex=True, figsize=figsize)
    dgp.set_xlabel('Generations')
    rfp.set_ylabel('R-Factor')
    dgp.set_ylabel('Generations since last change')
    # plot markers
    part = 0
    for (i, g) in enumerate(gens):
        if g > gens[-1] * 0.2:
            part = len(gens) - i
            break
    part = max(100, part)
    rfmin, rfmax = min(rfacsMin[-part:]), max(rfacsMax[-part:])
    if rfac_predict:
        (pred_x, pred_y) = tuple(zip(*rfac_predict))
        lowbound = max(min(min(pred_y[int(len(pred_y)*0.5):]), rfmin), 0)
    else:
        lowbound = rfmin
    rYrange = [lowbound-(rfmax-lowbound)*0.1, rfmax+(rfmax-rfmin)*0.1]
    if abs(rYrange[1] - rYrange[0]) < 1e-5:
        same = rYrange[0]
        rYrange[0] -= 0.05*same
        rYrange[1] += 0.05*same
    labely = rYrange[0] + (rYrange[1]-rYrange[0])*0.99
    xoff = gens[-1]*0.005
    for (xpos, label) in markers:
        rfp.axvline(x=xpos, lw=0.5, c="black")
        dgp.axvline(x=xpos, lw=0.5, c="black")
        rfp.text(xpos + xoff, labely, label, rotation=-90,
                 verticalalignment="top", size=4)
    # plot data
    if rfac_predict:
        rfp.step(pred_x, pred_y, where="post", color="seagreen",
                 label="Prediction")
    rfp.plot(gens, rfacsMin, '-', color='black', label="Best")
    rfp.fill_between(gens, rfacsMin, rfacsMax, facecolor='grey',
                     alpha=0.2, label="Range")
    rfp.plot(gens, rfacsMean, label="Mean")
    (x, y, s, c) = list(zip(*rp.rfacscatter))
    rfp.scatter(x, y, s=s, c=c)
    scatcol = {"all": "black", "dec": "tab:blue", "best": "tab:red"}
    labels = {"all": "Changes to any", "dec": "Changes in best 10%",
              "best": "Changes to best"}
    sizes = {"all": 4, "dec": 5, "best": 6}
    maxy = 1
    for k in ["all", "dec", "best"]:
        x, y = deltagens[k].keys(), deltagens[k].values()
        dgp.scatter(x, y, s=sizes[k], c=scatcol[k])
        maxdgx, maxdgy = [1], [0]
        for g in deltagens[k].keys():
            if deltagens[k][g] >= maxdgy[-1]:
                maxdgx.append(g)
                maxdgy.append(deltagens[k][g])
        if gens[-1] - max(deltagens[k].keys()) >= maxdgy[-1]:
            maxdgx.append(gens[-1])
            maxdgy.append(gens[-1] - max(deltagens[k].keys()))
        dgp.plot(maxdgx, maxdgy, '-', c=scatcol[k], label=labels[k])
        maxy = max(maxy, max(maxdgy), max(y))
    # layout
    rfp.set_ylim(rYrange)
    dgp.set_ylim([0, maxy*1.1])
    dgp.ticklabel_format(axis="x", style="sci", scilimits=(0, 4))
    fig.tight_layout()
    rfp.legend(loc="lower left")
    dgp.legend(loc="upper left")
    figs.append(fig)

    # SEARCH PARAMETER SCATTER
    fig, axs = plt.subplots(figsPerPage, figsize=figsize, squeeze=True)
    figcount = 0
    labels = {"geo": "GEOMETRY", "vib": "VIBRATION",
              "occ": "OCCUPATION", "dom": "DOMAIN AREAS"}
    offsets = []
    rpToDo = [rp]
    if rp.domainParams:
        rpToDo.extend([dp.rp for dp in rp.domainParams])
    for (k, crp) in enumerate(rpToDo):
        sps = [sp for sp in crp.searchpars if sp.el != "vac" and sp.steps > 1]
        if not rp.domainParams:
            confindex = 0
        elif crp != rp:
            confindex = k-1
        for mode in ["dom", "geo", "vib", "occ"]:
            if rp.domainParams and crp == rp and mode != "dom":
                continue
            if not rp.domainParams and mode == "dom":
                continue
            spm = [sp for sp in sps if sp.mode == mode]
            while len(spm) > 0:
                if figcount >= figsPerPage:
                    fig.tight_layout()
                    figs.append(fig)
                    fig, axs = plt.subplots(figsPerPage, figsize=figsize,
                                            squeeze=True)
                    figcount = 0
                plotpars = spm[:parsPerFig]
                spm = spm[parsPerFig:]
                title = labels[mode]
                addinfo = []
                if mode != "dom" and rp.domainParams:
                    addinfo.append("domain {}"
                                   .format(rp.domainParams[k-1].name))
                if len(crp.disp_blocks) > 1:
                    addinfo.append("search {}".format(searchname[:20]))
                if addinfo:
                    title += " ("
                    while addinfo:
                        title += addinfo.pop(0)
                        if addinfo:
                            title += ", "
                        else:
                            title += ")"
                axs[figcount].set_title(title, pad=8)
                pltpoints = []  # x, y, color, size, alpha
                bestpoints = []
                xlabels = []
                predict = []
                for (i, par) in enumerate(plotpars):
                    vals = []
                    for (j, conf) in enumerate(lastconfig):
                        if mode != "dom":
                            val = ((conf[confindex][1][crp.searchpars
                                                       .index(par)]-1)
                                   / (par.steps-1))
                        else:
                            val = conf[i][0] / 100
                        r, g, b = allcolors[j]
                        if par.linkedTo is None and par.restrictTo is None:
                            alpha = 1.0
                            vals.append(val)
                        else:
                            alpha = 0.5
                        color = (r, g, b, alpha)
                        pltpoints.append((i+1, val, color, 1))
                        if j == 0:
                            bestpoints.append((i+1, val, alpha))
                    if mode == "dom":
                        xlabels.append("#{}\n{}".format(
                            i+1, rp.domainParams[i].name))
                        edgetext = ["0%", "100%"]
                    else:
                        if mode != "occ":
                            el = par.el
                        else:
                            el = par.atom.el
                        xlabels.append(f'#{par.atom.num}\n{el}')
                        edgetext = ["", ""]
                        if isinstance(par.edges[0], (np.floating, float)):
                            edgetext = [str(round(v, 4)) for v in par.edges]
                        elif type(par.edges[0]) == np.ndarray:
                            edgetext = ["[" + ", ".join([str(round(f, 4))
                                                         for f in (
                                                    v * np.array([1, 1, -1]))])
                                        + "]" for v in par.edges]
                        elif type(par.edges[0]) == str:
                            edgetext = par.edges
                    axs[figcount].text(i+1, -0.03, edgetext[0], fontsize=3,
                                       ha="center", va="top")
                    axs[figcount].text(i+1, 1.02, edgetext[1], fontsize=3,
                                       ha="center", va="bottom")
                    if vals:
                        offsets.append(np.std(vals))
                        # mean = np.mean(vals)
                        # offsets.append(np.mean([abs(v - mean) for v in vals])
                        #                * 2)
                    check = ("err_co", "err_unco")
                    if (par.parabolaFit["min"] is not None and
                            not all(np.isnan(par.parabolaFit[s])
                                    for s in check)):
                        # alpha = 0.5
                        # if par.linkedTo is None and par.rstrictTo is None:
                        #     alpha = 1.0
                        a = alpha
                        n = (par.steps - 1)
                        good_fit = True
                        if not any(any(0 < ((par.parabolaFit["min"] - 1)/n
                                            + sign*par.parabolaFit[s]/n) < 1
                                       for sign in (+1, -1)) for s in check):
                            good_fit = False
                            a = 0.5
                        err_unco = par.parabolaFit["err_unco"] / n
                        err_co = par.parabolaFit["err_co"] / n
                        if np.isnan(err_unco):
                            err_unco = 2
                        if np.isnan(err_co):
                            err_co = 2
                        predict.append((i+1,
                                        (par.parabolaFit["min"] - 1) / n,
                                        err_unco,
                                        err_co,
                                        a, good_fit))

                # combine duplicates:
                i = 0
                while i < len(pltpoints):
                    x, y, c, s = pltpoints[i]
                    j = i+1
                    while j < len(pltpoints):
                        if pltpoints[j][0] == x and pltpoints[j][1] == y:
                            s += pltpoints[j][3]
                            if pltpoints[j][2][0] < c[0]:
                                # use the color closer to black
                                c = (pltpoints[j][2][0], c[1], c[2], c[3])
                            pltpoints.pop(j)
                        else:
                            j += 1
                    pltpoints[i] = (x, y, c, s)
                    i += 1
                if predict:
                    m = MarkerStyle("D")
                    m._transform.scale(1.0, 0.5)
                    err_off = 0.08  # error bar offset
                    for alpha in set([p[4] for p in predict]):
                        pred_ok = [p for p in predict
                                   if p[5] and p[4] == alpha]
                        pl1 = tuple(zip(*pred_ok))
                        pl2 = tuple(zip(*[p for p in predict
                                          if p not in pred_ok
                                          and p[4] == alpha]))
                        for plx, c in ((pl1, "seagreen"), (pl2, "red")):
                            if not plx:
                                continue
                            axs[figcount].scatter(plx[0], plx[1], color=c,
                                                  alpha=alpha, marker=m, s=50)
                        # errors
                        for ind in (2, 3):
                            pred_ok = [p for p in predict if
                                       (not np.isnan(p[ind]) and p[ind] < 1)
                                       and p[5] and p[4] == alpha
                                       and any(0 < p[1] + sign*p[ind] < 1
                                               for sign in (+1, -1))]
                            pl1 = tuple(zip(*pred_ok))
                            pl2 = tuple(zip(*[p for p in predict
                                              if p not in pred_ok
                                              and p[5] and p[4] == alpha]))
                            for plx, c in ((pl1, "seagreen"), (pl2, "red")):
                                if not plx:
                                    continue
                                off = -err_off if ind == 2 else err_off
                                px = np.array(plx[0], dtype=float) + off
                                axs[figcount].errorbar(
                                    px, plx[1], plx[ind], capsize=2,
                                    fmt="none", color=c, alpha=alpha,
                                    lw=0.5)
                x, y = [p[0] for p in pltpoints], [p[1] for p in pltpoints]
                c, s = [p[2] for p in pltpoints], [p[3] for p in pltpoints]
                axs[figcount].plot([0, parsPerFig+2], [0.5, 0.5], color='grey',
                                   alpha=0.2)
                axs[figcount].scatter(x, y, s=s, c=c)
                for (px, py, alpha) in bestpoints:
                    axs[figcount].annotate("", (px+0.05, py), (px+0.25, py),
                                           arrowprops=dict(arrowstyle="wedge",
                                                           facecolor="black",
                                                           alpha=alpha))
                    axs[figcount].annotate("", (px-0.05, py), (px-0.25, py),
                                           arrowprops=dict(arrowstyle="wedge",
                                                           facecolor="black",
                                                           alpha=alpha))
                xlabels.extend([""] * (parsPerFig - len(xlabels)))
                axs[figcount].set_xlim([0, parsPerFig+1])
                axs[figcount].set_ylim([0, 1])
                axs[figcount].set_xticks(list(range(1, parsPerFig+1)))
                axs[figcount].set_xticklabels(xlabels)
                axs[figcount].tick_params(which='both', top=False,
                                          bottom=False, left=False,
                                          right=False, labelleft=False,
                                          labelbottom=True, pad=4)
                figcount += 1
    if offsets:
        rp.parScatter[-1].append((gens[-1], np.mean(offsets), max(offsets)))
    for i in range(figcount, figsPerPage):
        axs[i].axis('off')
    if fig not in figs:
        fig.tight_layout()
        figs.append(fig)

    # save
    if searchname in rp.lastParScatterFigs:
        for f in rp.lastParScatterFigs[searchname]:
            try:
                plt.close(f)
            except Exception:
                pass
    rp.lastParScatterFigs[searchname] = figs[1:]
    try:
        pdf = PdfPages(outname)
        for fig in figs:
            pdf.savefig(fig)
    except PermissionError:
        logger.warning("Failed to write to " + outname
                       + ": Permission denied.")
    except KeyboardInterrupt:
        raise
    except Exception:
        logger.warning("Failed to write to "+outname)
        raise
    finally:
        try:
            pdf.close()
        except Exception:
            pass
    for fig in [f for f in figs if
                searchname not in rp.lastParScatterFigs or
                f not in rp.lastParScatterFigs[searchname]]:
        try:
            plt.close(fig)
        except Exception:
            pass
    return None


def writeSearchReportPdf(rp, outname="Search-report.pdf"):
    """
    Writes a pdf file with reports on R-factor convergence and parameter
    scatter, collated over the entire run (i.e. potentially multiple searches).

    Parameters
    ----------
    rp : Rparams
        The run parameters
    outname : str, optional
        The file name to write to. The default is "Search-report.pdf".

    Returns
    -------
    None.

    """
    global _CAN_PLOT
    if not _CAN_PLOT:
        return None
    allmin = []
    allmax = []
    allmean = []
    allgens = []
    markers = []
    parScatterLines = []  # list of lists [gens, mean, max] per search
    gencount = 0
    for i in range(0, len(rp.searchplots)):
        (name, gens, rmin, rmax, rmean) = rp.searchplots[i]
        markers.append((gencount, "Search "+name))
        allgens.extend([v + gencount for v in gens])
        allmin.extend(rmin)
        allmax.extend(rmax)
        allmean.extend(rmean)
        if rp.parScatter[i]:
            parScatterLines.append(list(zip(*rp.parScatter[i])))
            parScatterLines[-1][0] = [v + gencount for v in
                                      parScatterLines[-1][0]]
        gencount = allgens[-1]

    figsize = (5.8, 8.3)
    figs = []
    # R-FACTORS AND MEAN SCATTER
    # create figure
    fig, (rfp, msp) = plt.subplots(2, 1, sharex=True, figsize=figsize)
    msp.set_xlabel('Generations')
    rfp.set_ylabel('R-Factor')
    msp.set_ylabel('Parameter scatter')
    # plot markers
    part = 0
    for (i, g) in enumerate(allgens):
        if g > allgens[-1] * 0.2:
            part = len(allgens) - i
            break
    part = max(50, part)

    # Notice that we take max(mean) rather than max(max) in order
    # to zoom-in in the interesting region. Using max(max) would
    # normally obscure this region, as the worst R could be
    # seriously worse than the mean.
    rfmin, rfmax = min(allmin[-part:]), max(allmean[-part:])
    if rfmax <= rfmin:
        rfmin *= 0.95
        rfmax *= 1.05
    rYrange = [rfmin-(rfmax-rfmin)*0.1, rfmax+(rfmax-rfmin)*0.1]
    rYrange[1] = max(rYrange[1], max(allmin)+(rfmax-rfmin)*0.1)

    labely = rYrange[0] + (rYrange[1]-rYrange[0])*0.99
    xoff = allgens[-1]*0.005
    for (xpos, label) in markers:
        rfp.axvline(x=xpos, lw=0.5, c="black")
        msp.axvline(x=xpos, lw=0.5, c="black")
        rfp.text(xpos + xoff, labely, label, rotation=-90,
                 verticalalignment="top", size=4)
    # plot data

    rfp.fill_between(allgens, allmin, allmax, facecolor='grey',
                     alpha=0.2, label="Range")
    rfp.plot(allgens, allmean, label="Mean")
    rfp.plot(allgens, allmin, '-', color='black', label="Best")

    labelled = False
    scattermax = 0
    for (allgens, psmean, psmax) in parScatterLines:
        meanline, = msp.plot(allgens, psmean, '-', color="tab:blue")
        maxline, = msp.plot(allgens, psmax, '-', color="black")
        scattermax = max(scattermax, max(psmax))
        if not labelled:
            meanline.set_label('Mean parameter \u03C3')    # sigma
            maxline.set_label('Highest parameter \u03C3')  # sigma
            labelled = True
    # Avoid matplotlib warnings if scattermax remained == 0. This
    # simply means that the all the individuals in the population
    # converged to the same configuration.
    if scattermax <= 1e-5:
        scattermax = 0.05

    # layout
    rfp.set_ylim(rYrange)
    msp.set_ylim([0, scattermax*1.1])
    msp.ticklabel_format(axis="x", style="sci", scilimits=(0, 4))
    fig.tight_layout()
    rfp.legend(loc="lower left")
    msp.legend(loc="lower left")

    # ADD SCATTERS
    for k in rp.lastParScatterFigs.keys():
        figs.append(rp.lastParScatterFigs[k])
    # save
    try:
        pdf = PdfPages(outname)
        pdf.savefig(fig)
        for k in rp.lastParScatterFigs.keys():
            for f in rp.lastParScatterFigs[k]:
                pdf.savefig(f)
    except PermissionError:
        logger.warning("Failed to write to " + outname
                       + ": Permission denied.")
    except KeyboardInterrupt:
        raise
    except Exception:
        logger.warning("Failed to write to "+outname)
        raise
    finally:
        try:
            pdf.close()
        except Exception:
            pass
    try:
        plt.close(fig)
    except Exception:
        pass
    return None



"""
======================================
  ViPErLEED Graphical User Interface
======================================
   *** Module guilib.widgetslib ***

Library of functions that are common to several Qt objects

Created: 2020-01-12
Author: Michele Riva
"""

import inspect
import re

import PyQt5.QtGui as qtg
import PyQt5.QtWidgets as qtw

from viperleed import guilib as gl
# import guilib as gl


################################################################################
#                                   FUNCTIONS                                  #
################################################################################


def editStyleSheet(qwidget, new_entries):
    """
    Append or replace new_entries in the styleSheet of qwidget
    
    Parameters
    ----------
    qwidget: QWidget
             QWidget whose style sheet will be edited
    new_entries: str
                 list of properties and values that are to be inserted in the
                 style sheet.
                 Format: "property1: value1; property2: value2; ..."
    """
    if not isinstance(qwidget, qtw.QWidget):
        raise ValueError("Argument 0 of editStyleSheet must be a QWidget "
                         f"subclass. Found {type(qwidget)} instead")

    if not isinstance(new_entries, str):
        raise TypeError("Argument 1 of editStyleSheet must be a string. "
                        f"Found {type(new_entries)} instead")

    entry_re = r"""
        (?P<property>[-a-z ]+)              # property:
                                            #    lowercase letters, '-' and ' '  
        :                                   # colon
        (?P<value>[\w\%\#\(\),:.-_\\\/ ]+)  # value:
                                            #    alphanumeric, %, #,
                                            #    parentheses, ' ' and
                                            #    .,:-_ \/
        """
    checkFormat = re.match(r"".join((r'^(', entry_re, r';)+$')),
                           new_entries, re.VERBOSE)
    if checkFormat is None:
        raise ValueError("Argument 1 of editStyleSheet must be a string "
                         "of the form 'property1: value1; "
                         "property2: value2; ...'")

    if len(qwidget.styleSheet()) == 0: # no style sheet present
        qwidget.setStyleSheet(new_entries)
        return

    # prepare a compiled re that will be used on all entries
    entry_format = re.compile(entry_re, re.VERBOSE)

    new_entries = (entry
                   for entry in re.split(';', new_entries)
                   if len(entry) > 0)

    for entry in new_entries:
        e_match = entry_format.match(entry)
        property = e_match.group('property')
        value = e_match.group('value')
        # check whether the entry is already present
        if property in qwidget.styleSheet():
            # update what's after "property:" till ";" (included) with value
            start = qwidget.styleSheet().index(property) + len(property) + 1
            # +1 to skip the ':'
            end = qwidget.styleSheet().index(';', start) + 1
            old = qwidget.styleSheet()
            qwidget.setStyleSheet("".join((old[:start], value, old[end:])))
        else: # otherwise simply append the new property
            qwidget.setStyleSheet(qwidget.styleSheet() + entry)


def drawText(painter, text, transform=None, combine=False):
    # reimplementation of text painting function that does not give awful
    # output. type(painter)==QPainter, type(p)==QPointF, type(text)==QString

    rawFont = qtg.QRawFont.fromFont(painter.font())
    indexes = rawFont.glyphIndexesForString(text)
    
    painter.save();
    paths = [rawFont.pathForGlyph(index) for index in indexes]
    advances = rawFont.advancesForGlyphIndexes(indexes,
                                               qtg.QRawFont.UseDesignMetrics
                                               | qtg.QRawFont.KernedAdvances)
    if transform is not None:
        painter.setWorldTransform(transform, combine=combine)
    for (path,advance) in zip(paths,advances):
        painter.fillPath(path, painter.pen().brush())
        painter.translate(advance)
    painter.restore()

def get_all_children_widgets(parent, exclude_parent=False, recursive=True):
    """
    This is an extension of the QObject.children() method that finds
    all the children QWidgets of a QObject.

    Parameters
    ----------
    parent: QWidget or QLayout (or subclasses)
    exclude_parent: bool, default = False
                    If True parent is not contained in the returned set
    recursive: bool, default = True
               If False, only first-generation widgets are returned 
               If True, all-generations widgets are included

    Returns
    -------
    list of QWidget
        - if parent is a QWidget:
          All QWidgets that are children of any depth of parent
        - if parent is a QLayout:
          All the QWidgets managed by parent and all their children of any depth
    """
    if not isinstance(parent, (qtw.QWidget, qtw.QLayout)):
        return set()
    
    children = set(parent.children())
    
    # find all widgets that are directly children of parent
    childrenWidgs = {child for child in [*children, parent]
                      if isinstance(child, qtw.QWidget)}
    # find all layouts that are children of parent
    childrenLays = {child for child in [*children, parent]
                      if isinstance(child, qtw.QLayout)}

    # add the widgets that are managed by the layouts in the list of children
    for lay in childrenLays:
        to_add = (lay.itemAt(idx).widget()
                  for idx in range(lay.count())
                  if isinstance(lay.itemAt(idx), qtw.QWidgetItem))
        childrenWidgs.update(widg
                             for widg in to_add
                             if isinstance(widg, qtw.QWidget))
    
    if recursive:
        # and run recursively to find all the nested children
        for child in childrenWidgs.copy():
            if child != parent:
                childrenWidgs.update(get_all_children_widgets(child))
    
    if exclude_parent:
        childrenWidgs.discard(parent)
    
    return childrenWidgs


################################################################################
#                                   CLASSES                                    #
################################################################################


class AllGUIFonts():  ## > Will handle in a different way!
    # May be worth removing the __init__ completely since all instances
    # have anyway the same 'value' for all fonts
    def __init__(self):
        self.buttonFont = qtg.QFont()
        self.smallTextFont = qtg.QFont()
        self.labelFont = qtg.QFont()
        self.smallButtonFont = qtg.QFont()
        self.largeTextFont = qtg.QFont()
        self.plotTitleFont = qtg.QFont()
        
        allfonts = [self.buttonFont,
                    self.smallTextFont,
                    self.labelFont,
                    self.smallButtonFont,
                    self.largeTextFont,
                    self.plotTitleFont]
        
        [x.setFamily("DejaVu Sans") for x in allfonts]
        
        self.buttonFont.setPointSize(10)
        self.smallTextFont.setPointSize(9)
        self.labelFont.setPointSize(10)
        self.smallButtonFont.setPointSize(8)
        self.largeTextFont.setPointSize(12)
        self.plotTitleFont.setPointSize(14)
        
        self.mathFont = qtg.QFont()
        self.mathFont.setFamily("CMU Serif")
        self.mathFont.setPointSize(15)
        self.mathFont.setStyleStrategy(qtg.QFont.PreferAntialias)



"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes.symdomains ***

Defines the LEEDSymetryDomains class, used for constructing a LEED pattern

Author: Michele Riva
Created: 2021-03-13
"""

from collections.abc import Sequence
import itertools

import numpy as np

from viperleed import guilib as gl


class LEEDSymmetryDomains(Sequence):
    """
    Collection of LEED domains generated from one superlattice matrix
    and the symmetry operations of a bulk group. LEEDSymmetryDomains is an
    immutable sequence (i.e., tuple-like). Calling LEEDSymmetryDomains[i]
    returns a viperleed.Lattice of the domain
    """
    # @gl.exec_time
    def __init__(self, leed_parameters):
        self.__parameters = gl.LEEDParameters(leed_parameters)
        super().__init__()
        
        # self.__operations is a list of the bulk symmetry operations generating
        # distinct domains. Read by self.operations
        self.__operations = self.__find_domain_operations()
        
        # self.__superlattices is a numpy.ndarray of the superlattice matrices
        # that generate distinct domains. It can be accessed via the attribute
        # self.superlattices, and it is permanently set there the first time the
        # attribute is accessed
        self.__superlattices = None
        
        # The return value of self.equivalent_spots has beams expressed as
        # tuples, but including ONLY THE NUMERATOR of the beam indices with
        # respect to the bulk. Use self.denominator_for_bulk_beams to retrieve
        # the correct denominator as an integer, which accesses the private
        # property self.__denominator. self.__denominator is set once and for
        # all the first time self.denominator_for_bulk_beams is called
        self.__denominator = None

        # self.__domains is the underlying list that is accessed when issuing
        # self[index]. It's a list of gl.Lattice instances
        self.__domains = self.__build_domains()

        # self.__equiv_spots_no_superpos is a list of dictionaries of
        # dictionaries, one element per domain, ordered as in self.__domains
        # (i.e., in self).
        # The outer dictionary is labeled by 'norm', 'other', or an azimuthal
        # angle. For each of the angle settings, the inner dictionary is in the
        # form {beam_i: set(beams_i)} with beams of domain i and their
        # equivalents within the domain at the specified primary beam angle(s).
        # Used only internally, cannot be accessed
        self.__equiv_spots_no_superpos = self.__get_spot_equivalence()

        # self.__extinct_spots list of dictionaries, one element per domain,
        # same order as in self.__domains (i.e., in self). Each dictionary has
        # keys 'norm', 'other', or an azimuthal angle, and as values a numpy
        # array that lists the extinct beams of the domain at the specified
        # beam angle(s).
        # Used only internally, cannot be accessed
        self.__extinct_spots = self.__get_extinct()

        # self.__equiv_spots_no_superpos and self.__extinct_spots are used in
        # the public method self.equivalent_spots(domains, theta, phi) for
        # creating the appropriate LEEDEquivalentBeams instance (unless it's
        # already cached), whose reference is stored into self.__last_eq.
        # Used only internally, cannot be accessed
        self.__last_eq = None
    
    def __repr__(self):
        txt = (f"{self.cell_shape} "
               + f"viperleed.LEEDSymmetryDomains({self.__parameters})")
        return txt
    
    def __len__(self):
        return len(self.__domains)
    
    def __getitem__(self, el):
        # NB: if a slice is given, the result is NOT type-preserving, i.e.,
        # it will not be a LEEDSymmetryDomains but a list of Lattice(s)         # Will this be a problem?
        return self.__domains[el]

    @property
    def groups(self):
        return {'surf': self.__parameters['surfGroup'],
                'bulk': self.__parameters['bulkGroup']}

    @property
    def bulk_basis(self):
        """
        Returns the reciprocal-space basis of the bulk lattice
        """
        return np.dot(self.superlattices[0].T, self[0].basis).round(10)

    @property
    def n_domains(self):
        """
        Number of distinct domains produced by the symmetry operations of the
        bulk
        """
        return len(self)

    @property
    def operations(self):
        return self.__operations

    @property
    def superlattices(self):
        """
        Numpy ndarray of superlattice matrices generating the symmetry-related
        domains. The first element is the one generating the lattice whose basis
        is given in the constructor as a LEED parameter
        """
        if self.__superlattices is None:
            superlattice = self.__parameters['SUPERLATTICE']
            self.__superlattices = np.einsum('ij,mjk->mik',
                                             superlattice,
                                             self.operations)
        return self.__superlattices

    @property
    def g_vectors(self):
        """
        g_vectors[i] is a list of the lattice vectors of domain i in the same
        coordinate system as the bulk basis
        """
        return [dom.lattice for dom in self]

    @property
    def denominator_for_bulk_beams(self):
        if self.__denominator is None:
            self.__denominator = int(
                abs(round(np.linalg.det(self.superlattices[0])))
            )
        return self.__denominator

    def angular_offsets(self, zero_pi=False):
        """
        Returns a list of the angles between the first lattice vectors of all
        domains and those of the first domain
        """
        first_dom_angle = gl.orientation(self[0].real_basis[0], zero_pi)
        return [gl.orientation(dom.real_basis[0], zero_pi) - first_dom_angle
                for dom in self]

    # @gl.profile_lines
    def beams_equivalent_to(self, beam, **kwargs):    # this will probably move to LEEDPattern later on? As is, it's fast enough not to bother
        """
        Given a beam, returns beams that are equivalent to it. This function
        cannot be used before issuing once self.equivalent_spots at the same
        angular conditions and with the same domains.
        
        Parameters
        ----------
        beam : 2-element array-like of numbers
            Some reference to the beam. Which reference is passed should be
            explicitly specified with the in_format parameter
        in_format : str, keyword only
            Acceptable values are:
                - 'fractional':
                    when passing the full bulk fractional indices. In this case,
                    beam can be a list-like with two int, or two floats, or a
                    string in the form 'num1/den1, num2/den2' ('/den_i' can be
                    skipped if == 1)
                - 'numerator':
                    when passing only the numerator of the bulk fractional index
                    In this case, beam should be a tuple of integers. The
                    denominator is taken automatically from the superlattice
                    matrix
                - 'g':
                    when passing the vector position in Cartesian coordinates.
                    In this case, beam should be expressed in a reference system
                    with z orthogonal to the surface (DO WE NEED THIS? PROBABLY
                    NOT AFTER I FIGURE OUT THE GENERAL POSITIONS ON THE SCREEN),
                    and in the same in-plane reference as the bulk basis
        - out_format='' : str, optional (default out_format=in_format)
            Same values as in_format. 'numerator' probably doesn't make much
            sense.

        Returns
        -------
        list
            as many elements as there are beams equivalent to beam, each one is
            a (beam, index, text) tuple, with
            beam : tuple
                formatted as requested in out_format
            index : gl.BeamIndex
                fractional indices
            text : str
                form "i+(j)+..." where i,j,... are the indices of the domains
                that contribute to the beam, and parenthesized ones are extinct
                contributions
        """
        # Potentially useful for the hovering annotations.
        # They need the following info:
        # - beams equivalent to beam
        # - for all beams:
        #   * some type of coordinate, probably best to give out the g vectors?
        #     it will need to be processed later on to get the right rotation
        #     (depending on the view angle), and, consequently the pixel
        #     coordinates in the canvas
        #   * the fractional index + a list of the domains that contribute to
        #     that beam, perhaps already formatted as a string?
        if self.__last_eq is None:
            raise RuntimeError("beams_equivalent_to cannot be used before "
                               "running once self.equivalent_spots at the same "
                               "angular conditions and with the same domains")
        in_format = kwargs.get('in_format', '')
        if in_format not in ('fractional', 'numerator', 'g'):
            raise ValueError("beams_equivalent_to: in_format is mandatory, and "
                             "should be either 'fractional', 'numerator', or "
                             "'g'")
        elif in_format == 'g':
            # Since g_bulk = (h,k)_bulk @ basis_bulk, with basis_bulk the
            # reciprocal-lattice bulk basis, then
            # (h,k)_bulk = g_bulk @ basis_bulk^(-1)
            # The indices are the fractional ones in this case.
            beam_indices = np.dot(beam, np.linalg.inv(self.bulk_basis))
        else:
            beam_indices = beam

        # Now process the beam indices to have only the numerator, as this is
        # what is used in the dictionaries
        if in_format in ('fractional', 'g'):
            beam_indices = [b*self.denominator_for_bulk_beams
                            for b in beam_indices]
            # check that the indices are consistent with the bulk basis
            if any(abs(round(b) - b) > 1e-4 for b in beam_indices):
                raise ValueError("beams_equivalent_to: beam indices are "
                                 "incompatible with the bulk basis")
            beam_indices = [int(round(b)) for b in beam_indices]

        # make the indices a tuple for lookup in the dictionary of equivalent
        # beams
        beam_indices = tuple(beam_indices)
        try: 
            eq_beams = self.__last_eq.equivalent_beams_dict[beam_indices]
        except KeyError:
            raise ValueError(f"beams_equivalent_to: beam {beam} (transformed to"
                             f" {beam_indices}) not found. "
                             "It may not be an acceptable index for the "
                             "system, or may lie outside the energy range")

        # Transform eq_beams into a list, as it's a set in the dictionary, but
        # we need to keep the order for the output
        eq_beams = list(eq_beams)

        # Now get the names of the domains that overlap at each of the beams,
        # also accounting for whether the domain contributes with an extinct
        # spot
        beams, _, domains, extinct_domains = zip(*self.__last_eq.indexed_beams)
        overlapping = []
        for b in eq_beams:
            idx = beams.index(b)
            overlap = [f"({d})" if d in extinct_domains[idx] else f"{d}"
                       for d in domains[idx]]
            overlapping.append("+".join(overlap))

        # Finally process the beam indices according to the format requested
        fractional = [gl.BeamIndex(b,
                                   denominator=self.denominator_for_bulk_beams,
                                   from_numerators=True)
                      for b in eq_beams]

        out_format = kwargs.get('out_format', in_format)
        if out_format == 'fractional':
            out_beams = fractional
        elif out_format == 'numerator':
            out_beams = [b for b in eq_beams]
        elif out_format == 'g':
            out_beams = np.dot(fractional, self.bulk_basis).astype(float)
        else:
            raise ValueError("beams_equivalent_to: invalid out_format "
                             f"{out_format!r}. Expected 'fractional', "
                             "'numerator', or 'g'")
        return list(zip(out_beams, fractional, overlapping))

    # @gl.profile_lines
    def equivalent_spots(self, domains=None, theta=None, phi=None):
        """
        Returns a list of beams grouped by equivalence, given a primary beam
        direction defined by polar and azimuthal angles
        
        Parameters
        ----------
        domains : list of int, or None (default=None)
            only the beams of the domains selected by these indices will be
            output. If None, all the domains are used.
        theta, phi : number or None
            polar and azimuthal directions in degrees of the primary beam. If
            not given or None, those defined at instantiation are used
            theta should be in the [-90, 90] range
            phi is positive counterclockwise, and measured from the x axis
                in the Cartesian reference of the real-space basis of the
                first domain
        """
        # type- and value-check the input, and processing when needed
        if theta is None:
            theta = self.__parameters['beamIncidence'][0]
        if phi is None:
            phi = self.__parameters['beamIncidence'][1]
        if not all(isinstance(angle, (int, float)) for angle in (theta, phi)):
            raise TypeError("LEEDSymmetryDomains: invalid angles. "
                            "Expected a real number.")
        theta, phi = gl.conventional_angles(theta, phi)

        if domains is None:
            domains = range(self.n_domains)
        elif not all(isinstance(dom, int) for dom in domains):
            raise ValueError("Invalid domain index. "
                             "All indices should be integers")
        elif any(dom < 0 or dom >= self.n_domains for dom in domains):
            raise ValueError("Domain index out of range. "
                             "Indices should be between 0 "
                             f"and {self.n_domains}.")

        # select the bare dictionaries
        kwargs = {'domains': domains, 'theta': theta, 'phi': phi}
        domains_dicts = self.__beams_for_primary_angles('equivalent', **kwargs)
        extinct_beams = self.__beams_for_primary_angles('extinct', **kwargs)
        
        # Now, only if the beam is impinging normally, one has to rework a bit
        # the dictionary. In fact, at normal incidence, one should not only
        # consider the spot equivalence within each domain, but also the
        # equivalence among different domains. For example, the two (2x1)
        # domains on a p4xx square lattice produce spots (1/2 0) and (0 1/2),
        # respectively, that are equivalent to one another at normal incidence.
        # This means that, at normal incidence, for each domain i, all the
        # beam "lists" (values) associated with beam j (key) are the same,
        # although beam j is a different tuple, and comprise all the beams
        # for all domains that are equivalent to the j-th beam of each domain
        #
        # Will construct a list of sets, in the same order as the keys,
        # where each set is the union of the stars of the domains. This is
        # possible since all dictionaries were created with the same order
        # of insertion (actually, from the same dictionary)
        if abs(theta) < 1e-4:
            all_stars = [set.union(*stars)
                         for stars in zip(*(d.values() for d in domains_dicts))]
            # now place the unions in the domain dictionaries, preserving the
            # original keys
            domains_dicts = [dict(zip(d.keys(), all_stars))
                             for d in domains_dicts]

        # Finally defer the processing to the LEEDEquivalentBeams class, that
        # needs to know the reciprocal-space bulk basis for processing.
        # Prepare the keyword arguments before:
        kwargs = {'extinct_lists': extinct_beams, 'basis': self.bulk_basis,
                  # currently using 1-based domain indices for display purposes
                  # as it may be easier to look at from the viewpoint of the
                  # final user
                  'domain_ids': [dom + 1 for dom in domains],
                  'superlattice': self.superlattices[0],
                  'angle_key': self.__key_from_angles(theta, phi, domains)}
        self.__last_eq = gl.LEEDEquivalentBeams(domains_dicts, **kwargs)
        return self.__last_eq.indexed_beams

    def __beams_for_primary_angles(self, which_beams, theta=None, phi=None,
                                   domains=None):
        """
        Returns the correct equivalent or extinct beams given the direction
        angles of the primary beam

        Parameters
        ----------
        which_beams : str
            Can be either 'equivalent' (or any contraction down to 'eq') or
            'extinct' (or any contraction down to 'ex'). Determines what is
            returned, the dictionary of equivalent beams or the lists of
            extinct ones, respectively
        theta, phi : number or None (default=None)
            if None, the angle given as a parameter in the constructor is used.
            The azimuthal angle phi is measured with respect to the x axis
            (positive counterclockwise), in the same Cartesian coordinate frame
            as the real-space basis of the first domain. Angles are in degrees.
        domains : list of int or None (default=None)
            select which domains should be output. If None, all domains are used

        Returns
        -------
        list of dict
            if which_beams[:2] == 'eq'
            keys : tuple
                beam j of domain i
            values : set of tuple
                the star of beam j in domain i
        list
            if which_beams[:2] == 'ex'
            
        """
        # check and fix the input if needed
        which_beams = which_beams[:2]
        if which_beams not in ('eq', 'ex'):
            raise ValueError("LEEDSymmetryDomains: invalid parameter for "
                             "selecting which beams to output. Must be 'eq'"
                             "or 'ex' for equivalent or extinct")
        if theta is None:
            theta = self.__parameters['beamIncidence'][0]
        if phi is None:
            phi = self.__parameters['beamIncidence'][1]

        if domains is None:
            domains = range(self.n_domains)

        # now select which keys need to be retrieved
        key = self.__key_from_angles(theta, phi, domains)

        # and from which dictionary
        if which_beams == 'eq':
            beams = self.__equiv_spots_no_superpos
        else:
            beams = self.__extinct_spots

        # retrieve the correct dictionary keys, falling back on 'other' if the
        # angle phi is not one of the special directions
        return [beams[dom].get(key, beams[dom]['other']) for dom in domains]

    def __key_from_angles(self, theta, phi, domains):
        """
        Given the polar and azimuthal angles of incidence of the beam,
        figure out which among the keys 'norm', 'other', or any of the mirror
        direction angles needs to be retrieved
        """
        a_angle = gl.orientation(self[0].real_basis[0], zero_pi=False)
        phi = (phi - a_angle) % 180

        if theta < 1e-4:
            key = 'norm'
        elif abs(phi - round(phi)) > 1e-4:
            # will need to use 'other' for all, as all angles in the
            # dictionaries are integers: 0, 30, 45, 60, ...
            key = 'other'
        else:
            key = str(round(phi))
        if any(key in self.__equiv_spots_no_superpos[dom] for dom in domains):
            return key
        return 'other'

    def __build_domains(self):
        """
        Returns the viperleed.Lattice(s) with self.superlattices matrices
        """
        # Create dummy surface lattice just to get the reciprocal basis
        surf = gl.Lattice(self.__parameters['surfBasis'])

        # and get the maximum screen radius
        max_radius = gl.screen_radius(self.__parameters['eMax'],
                                      self.__parameters['screenAperture'])

        # Then prepare the actual reciprocal lattice of the first domain
        domains = [gl.Lattice(surf.reciprocal_basis, space='reciprocal',
                              group=self.__parameters['surfGroup'],
                              limit=max_radius)]

        # and work out those of the others.
        # The reciprocal-space basis of domain i is:
        #    Bi = Mi^(-T) Bb
        # with Bb the bulk reciprocal basis, and Mi the superlattice matrix
        # of domain i, i.e.,
        #   Mi = M0 Gi
        # where Gi is the operation (of the bulk group) that creates domain i.
        # Thus
        #   Bi = (M0 Gi)^(-T) Bb = Gi^(-T) M0^(-T) Bb = Gi^(-T) B0
        # However, when transforming the basis, Gi needs to be expressed
        # in the coordinate system of the SURFACE, i.e.,
        #   Gi = Breal0 Gi_absolute Breal0^(-1)
        #      = Breal0 Brealb^(-1) Gi_bulk Brealb Breal0^(-1)
        #      = superlattice_0 Gi_bulk superlattice_0^(-1)
        superlattice = self.superlattices[0]
        inv_superlattice = np.linalg.inv(superlattice)
        ops_t = [np.linalg.multi_dot((superlattice,
                                      op,
                                      inv_superlattice))
                 for op in self.operations]
        ops = [np.linalg.inv(op).T for op in ops_t[1:]]  # skip E (1st domain)
        domains.extend([domains[0].transform(op, as_copy=True)
                        for op in ops])
        return domains

    def __find_domain_operations(self):
        """
        Finds the symmetry operations of the bulk lattice that give distinct
        domains in LEED.
        NB: this function gives different results than what one would get from
        LEEDPat, as we're interested also in the symmetry relations between the
        intensities of the spots, while LEEDPat cares only about the presence
        or not of any spot.

        Returns
        -------
        list of operations

        -------
        """
        # The current version is based on the concept of co-sets of a group.
        # Given a group G and a subgroup H, the left co-set of H with respect
        # to the group operation g of G is
        #           gH = {g*h : h in H}.
        # Additionally, we take into account that each element of G is found in
        # exactly only one co-set (e.g., H is the identity co-set).
        #
        # The operations g_i that generate distinct co-sets are those that will
        # generate distinct lattices

        # def __array_2_tuple(arr):
            # """
            # Convenience function that is used to convert a 2x2 array to a 2x2
            # tuple
            # """
            # return tuple(map(tuple, arr))

        # 1) project the operations from the surface group to the bulk; also
        #    rounding to integers
        project_to_bulk = np.linalg.inv(self.__parameters['SUPERLATTICE'])
        surf_ops = tuple(
            op.round().astype(int)
            for op
            in self.groups['surf'].transform(project_to_bulk)
            )

        # 2) keep track of the operations of the bulk group that are already in
        #    a co-set, and of the group operations giving distinct co-sets
        coset_ops = set()
        _ops = []

        # 3) run through the bulk operations, adding the operations of new
        #    co-sets to the set above
        bulk_ops = self.groups['bulk'].operations(include_3d=True)

        for bulk_op in bulk_ops:
            if bulk_op in coset_ops:
                continue
            _ops.append(bulk_op)
            coset = set(gl.two_by_two_array_to_tuple(np.dot(bulk_op, surf_op))
                        for surf_op in surf_ops)
            coset_ops.update(coset)

        return _ops

    def __get_spot_equivalence(self):
        """
        Determines which of the LEED beams of the domains belong to the same
        equivalence class under the symmetry operations of the domain itself,
        i.e., it determines the 'star' of each of the beams.
        Returns a list of dictionary of dictionaries, one list element per each
        domain. Each key in the outer dictionary is either 'norm', the angle
        in degrees of a mirror direction (with respect to the first real-space
        lattice vector of the first domain BETTER WRT BULK??), or 'other'.
        These keys are useful for selecting equivalent spots when the primary
        beam direction is, respectively, normal to the surface, contained in a
        mirror plane, or generic.

        Returns
        -------
        list of dicts, each one
            keys : str
                'norm' and 'other' keys are always present. In addition, there
                are as many keys as self.special_directions, each one is the
                angle in degrees between the mirror direction and the first
                real-space basis vector. The angles can only be a subset of:
                '0', '30', '45', '60', '90', '120', '135', '150'
            values : dict
                keys : tuple
                    one key per each index in self.hk
                values: set of tuples
                    all beam indices symmetry-equivalent to key, including key,
                    i.e., what's known as 'the star' of the beam in key
        """
        # Will create several dictionaries, labeled depending on the beam
        # incidence directions. There always will be one for 'norm' (normal)
        # incidence, as many as there are special_directions, each of
        # which will be indexed by the angle between the special direction
        # and the first basis vector, and one for 'other'. This makes it fast
        # to later select which of the equivalence relations to choose.
        # Since the beam equivalence dictionaries are the same for all domains,
        # work on the first one only , i.e., self[0], in integer-index notation,
        # and later on translate stuff to the other domains
        all_operations = self[0].group.operations()
        a_angle = gl.orientation(self[0].real_basis[0], zero_pi=False)
        
        # Prepare the dict keys:
        labels = ['norm', 'other']
        operations = [all_operations,        # Normal incidence -> all
                      (all_operations[0],)]  # Generic -> only identity
        # Now go through the mirror directions for the other keys.
        # For a beam along the mirror, the only operations that remain are
        # identity and the mirror itself
        for op, direction in zip(all_operations, self[0].special_directions):
            if direction is None:  # the operation is a rotation
                continue
            phi = gl.orientation(direction, zero_pi=False)
            labels.append(round((phi - a_angle) % 180))
            operations.append((all_operations[0], op))
        
        # Now use the labels and the operations found above to find the star of
        # each beam of the first domain in all cases
        beams_dict_first = {}
        for label, ops in zip(labels, operations):
            # Transform all the beams with all the operations
            hk_transformed = np.einsum('ilm,mj->jil', ops, self[0].hk.T)

            # convert all beams to tuples so they can be keys for the dictionary
            hk_tuples = gl.two_by_n_array_to_tuples(self[0].hk)
            
            # and convert to tuples also the transformed beams,
            # also using a set to remove duplicates
            beams_dict_first[label] = {
                k: set(gl.two_by_n_array_to_tuples(v, axis=1))
                for k, v in zip(hk_tuples, hk_transformed)
            }
        return self.__beams_dict_to_other_domains(beams_dict_first)
    
    def __get_extinct(self):
        """
        For each domain, finds the beams extinct due to glide symmetry, and
        populates dictionaries with the in-plane directions of the primary
        beam that makes them extinct
        """
        group = self.groups['surf'].group
        # Like for __get_spot_equivalence, do the calculation on the first
        # domain, then extend to the others by relabeling
        if 'g' not in group:
            # no glide, no need to bother
            return [{'norm': [], 'other': []}]*len(self)
        
        # get boolean arrays that pick the extinct spots along the [1 0] and
        # [0 1] directions (for the first domain, but they're the same for
        # the others)
        extinct_10 = (self[0].hk[:, 1] == 0) & (self[0].hk[:, 0] % 2 == 1)
        extinct_01 = (self[0].hk[:, 0] == 0) & (self[0].hk[:, 1] % 2 == 1)
        
        beams_dict_first = {'other': []}
        # pg, pmg, p4g are the only options, and exist only for square
        # or rectangular lattices, so the angles are only 0 and/or 90
        if '[1 0]' in group:
            beams_dict_first['norm'] = self[0].hk[extinct_10]
            beams_dict_first[0] = self[0].hk[extinct_10]
        elif '[0 1]' in group:
            beams_dict_first['norm'] = self[0].hk[extinct_01]
            beams_dict_first[90] = self[0].hk[extinct_01]
        else:
            beams_dict_first['norm'] = self[0].hk[extinct_10 | extinct_01]
            beams_dict_first[0] = self[0].hk[extinct_10]
            beams_dict_first[90] = self[0].hk[extinct_01]
        return self.__beams_dict_to_other_domains(beams_dict_first)

    def __beams_dict_to_other_domains(self, beams_dict_first):
        """
        Converts a beams dictionary from the first domain into a list of beams
        dictionaries for each domain. This is done by determining angular
        offsets between domains, restricting to 0...pi, and building the offset
        dictionaries. At the same time, we also convert the integer domain
        indices into fractional bulk indices via the superlattice matrices
        
        Parameters
        ----------
        beams_dict_first : dict
            keys : str or int
                can be 'norm', 'other' or integers with the angles of some
                special directions
            values : list-like or dict
                can either be a list of integer hk values of the first domain,
                or a dictionary of the form {hk0: {hk0, hk1, ...}, ...}

        Returns
        -------
        list of dict
            one list element for each of the domains, each dict has the same
            format as the one passed as parameters, except for the fact that
            keys are all strings, and angles are adjusted to all refer to the
            first lattice vector of the first domain. The contents of the
            dictionaries are the same as passed, but converted to be the
            numerator of the fractional indices with respect to the bulk
        """
        offsets = self.angular_offsets(zero_pi=True)
        beams_dicts_all_domains = []
        mu = self.denominator_for_bulk_beams
        for offset, superlattice in zip(offsets, self.superlattices):
            ddict = {}
            for label, beams in beams_dict_first.items():
                # set up the transformation matrix that gives fractional
                # indices in the bulk basis from the integer indices of
                # each domain
                transform = np.linalg.inv(superlattice).T

                # it is more convenient (faster) to still store only the
                # numerator of the indices as tuples, as this limits numerical
                # errors. The fractional indices can be retrieved when needed
                # by dividing by self.denominator_for_bulk_beams
                transform = (mu*transform).round().astype(int)
                bulk_beam_dict = self.__beams_to_bulk(beams, transform)
                if label in ['norm', 'other']:
                    ddict[label] = bulk_beam_dict
                else:
                    ddict[str(round((label + offset) % 180))] = bulk_beam_dict
            beams_dicts_all_domains.append(ddict.copy())
        return beams_dicts_all_domains

    def __beams_to_bulk(self, beams, transform):
        """
        Take either a list [hk0, hk1, ...] or a dict of the form
        {hk0: {hk1, hk2, ...}} with hk integer indices relative to one domain,
        and transform them to a list or a dict, respectively, with fractional
        indices all over.
        """
        # In case we have a dict, extract beams from the keys (one each key)
        # as well as those from the values, i.e., beams of the star of key
        if isinstance(beams, dict):
            k, v = zip(*beams.items())
            # pack them into a flat list, as this allows to run a dot product
            # in numpy that makes calculations much faster, despite the
            # additional work to unpack/re-pack stuff
            flattened_beams = list(itertools.chain(k, *v))
        elif len(beams) == 0:
            return beams
        else:
            flattened_beams = beams

        # transform the beams. For some reason, einsum is faster than dot
        transformed_beams = np.einsum('ij,jk', flattened_beams, transform)

        # Now transform them back to a list of tuples
        transformed_beams = list(gl.two_by_n_array_to_tuples(transformed_beams,
                                                             axis=1))
        if not isinstance(beams, dict):
            return transformed_beams

        # When a dict was passed, we need to split the 
        # beams once again into keys and values
        k_transf = transformed_beams[:len(k)]
        v_transf = transformed_beams[len(k):]

        # and figure out how long each of the values used to be, so they can
        # be split again correctly among the various keys
        split_idx = list(itertools.accumulate([0, *(len(i) for i in v)]))

        # Do the splitting, also converting again to set
        v_transf = [set(v_transf[i:j])
                    for i, j in zip(split_idx, split_idx[1:])]

        # and recreate the dictionary
        return dict(zip(k_transf, v_transf))
